# The command-line interface module creates an interface for
# interacting with the python program (wikicrawl). This is an implementation
# of the baker demo shown previously. The user can type in commands to
# make the program do things.

import baker
import sys

if sys.platform == 'linux':
    import readline # Needed for command history <up> and <down> arrows to work

from . import log
from . import model
from . import config

# Problem pages:
# Decision (from politics)
# Malaysia (goes inside parenthesis)
# Soft-sediment_deformation_structures (doesn't find link)
# Chemicals (loops at philosophical)

commander = baker.Baker()

class InteractiveInterface:
    def __init__(self):
        # Instantiate the variable self.model as an object
        # of instance of the Model class defined in the model
        # module. model.Model refers to the Model class in the
        # model module and this line creates a new variable (self.model)
        # which is a variable that is an instance of Model, i.e.
        # it has the type Model and has Model.methods() available
        # to it.
        #
        # self.model is a variable that is attached to the instance/object
        # returned by this constructor that has the type InteractiveInterface.
        self.model = model.Model()

    def run_command(self, args, main=True):
        """
        Runs the command-line interface for a single command.

        If called by InteractiveInterface.run(sys.argv), this method
        will execute the commands and arguments specified on command
        line when running this program. Alternatively, the code could
        pass in a different set of arguments to specify what to do.
        See start_command_loop() for more information.
        """
        try:
            commander.run(argv=args, main=True, help_on_error=True,
                          instance=self)
        except baker.CommandError as ex:
            log.LOGGER['cli'].warn('incorrect user input: %s' % ex)
            commander.usage()
        except baker.TopHelp as ex:
            commander.usage()
        except Exception as ex:
            log.LOGGER['cli'].error('caught general exception!!')
            log.LOGGER['cli'].error(ex)

    def start_command_loop(self):
        """
        Repeatedly asks the user what command to run until they exit.

        This method calls InteractiveInterface.run(args) a little bit
        differently. Instead of passing the arguments from the command-line
        that were passed in when invoking the python wikicrawl app,
        this asks the user for a line of textual input and passes
        those strings to run() as the arguments. This way, the user can
        access an interactive shell and repeatedly issue different
        commands while the application is running.
        """
        commander.usage()
        self.model.open_browser()
        while True:
            print('$ ', end = '') # Display to the user a command prompt
                                  # The dollar-sign is a common indication
                                  # of a shell that communicates to the user
                                  # that we are waiting for their textual
                                  # input. The end = '' indicates to python
                                  # to NOT drop to a newline after printing
                                  # in the terminal. Instead, let the user
                                  # type their command on the same line as
                                  # the printed '$ '.
            try:
                inp = input()
            except EOFError: # <ctrl>+D will send "End Line" and exit the command loop
                break

            # Note on "arguments" (mg):
            # Whenever a program is run in windows or *nix, the operating
            # system passes in the command string that was used to invoke
            # the program. You can append data in that command to configure
            # switches or values going into the program on the fly. For
            # example, you can invoke this wikicrawl app in more than one
            # way. You can of course run "python launcher.py" to run the
            # software but you can also pass in an argument. You can
            # alternatively run "python launcher.py <argument> <argument>..."
            # and the operating system will provide the <argument> values into
            # the process that is running as variables.
            #
            # In a real world use case, many commands provide switches to
            # adjust what the program does. For example,
            #
            # The command:
            #     find music -name "*justin*bieber*"
            # runs the "find" program and asks to find all the filenames that match the
            # pattern *justin*bieber* in the "music" directory.
            # (music, -name, "*justin*biever*") are argument parameters
            # that are passed into the program. The program is coded to
            # parse and interpret these values and execute differently based
            # on the values passed in. This is one way to pass in information
            # into a running program. Some other ways are to read from a file
            # (such as how we read from settings.py to load the runtime
            # configuration), from something called environment variables
            # (won't get into but another set of values provided to programs
            # from the operating system), or they can be hard-coded into
            # the application.
            #
            # Side note: arguments are not unique to python (almost all
            # programming languages implement arguments), the functionality
            # is defined by the application (some programs require arguments,
            # some are optional, and the syntax for sending in argument
            # parameters are different and defined by the individual programs,
            # and lastly, the first argument sent in is the script name or
            # filename of the script. In our case, the first argument is
            # the string "launcher.py". If the user invoked the command
            # as C:\Users\mguest\launcher.py then the first argument
            # would be C:\Users\mguest\launcher.py.

            # What this method (start_command_loop()) does is provide a
            # REPL shell which is a
            # read-eval-print-loop. It repeatedly asks the user for an
            # input (read), evaluates that input into an action (evaluate),
            # give the user some feedback (print), and start the process
            # over again (loop). When you call just "python", you are loading a
            # program that gives you a REPL interactive shell. The way
            # this wikicrawl app is implemented gives the user a REPL
            # that has commands to interact with wikipedia pages.

            # Because we take in the input as a single string, we do
            # a transformation to turn something like "do_random_page 5"
            # into ["launcher.py", "do_random_page", "5"] which is how
            # the arguments array would have been created if it were
            # passed in the initial command instead of typed and interpretted
            # as input as is done here.
            args = [sys.argv[0], ] + inp.split()

            # The user can at any point in the command pass the argument
            # switch "--help". If doing this, the command line interface
            # will instead print out the inline documentation associated
            # with this command and quit after doing so. For example,
            # the user can type "python launcher.py do_random_page --help"
            # and the program will spit out the generated documentation
            # for the do_random_page command and run nothing. In our case,
            # this documentation is created by the baker library and will
            # print out the docstring associated with the method. Try it
            # out in your shell (cmd.exe or powershell.exe) by invoking
            #     python launcher.py do_random_page --help
            # You will see the program spit out the heredoc below the
            # do_random_page method defined below.
            if '--help' in args:
                args.remove('--help')
                try:
                    print('command usage:')
                    commander.usage(args[1])
                except Exception as ex:
                    print(type(ex), ex)
                continue

            self.run_command(args, main=False)

    @commander.command
    def play_specific_page(self, title):
        pass

    @commander.command
    def play_random_page(self):
        """
        Instructs the wikicrawl application to play the game on a random
        article.
        """
        self.model.play_random_page()

    @commander.command
    def play_multiple(self, n):
        """
        Plays the wikicrawl game <n>-times.
        """
        try:
            n = int(n)
        except ValueError as ex:
            log.LOGGER['cli'].warn('failed to process "%s" as a parameter' % n)
            return False
        for i in range(n):
            self.model.play_random_page()

    @commander.command
    def exit(self):
        """
        Immediately exit the program.
        """
        sys.exit(0)

